#!/usr/bin/python

from patches import parse_patches
import os
import sys
import string
import glob
import bzrlib
from bzrlib.commands import Command
from bzrlib.branch import Branch
from bzrlib import DEFAULT_IGNORE
from hunk_selector import HunkSelector

DEFAULT_IGNORE.append('./.bzr-shelf*')

def run_bzr(args):
    if type(args) is str:
        args = [ args ]
    pipe = os.popen('bzr %s' % string.join(args, ' '), 'r')
    lines = pipe.readlines()
    if pipe.close() is not None:
        raise Exception("Failed running bzr")
    return lines

class QuitException(Exception):
    pass

class Shelf(object):
    def __init__(self):
        b = Branch.open_containing('.')[0]
        self.bzr_root = b.base

    def shelf_suffix(self, index):
        if index == 0:
            return ""
        else:
            return "-%d" % index

    def next_shelf(self):
        def name_sequence():
            i = 0
            while True:
                yield self.shelf_suffix(i)
                i = i + 1

        stem = os.path.join(self.bzr_root, '.bzr-shelf')
        for end in name_sequence():
            name = stem + end
            if not os.path.exists(name):
                return name

    def last_shelf(self):
        stem = os.path.join(self.bzr_root, '.bzr-shelf')
        shelves = glob.glob(stem)
        shelves.extend(glob.glob(stem + '-*'))
        def shelf_index(name):
            if name == stem:
                return 0
            return int(name[len(stem)+1:])
        shelvenums = [shelf_index(f) for f in shelves]
        shelvenums.sort()

        if len(shelvenums) == 0:
            return None
        return stem + self.shelf_suffix(shelvenums[-1])

    def get_shelf_message(self, shelf):
        prefix = "# shelf: "
        if not shelf.startswith(prefix):
            return None
        return shelf[len(prefix):shelf.index('\n')]

    def unshelve(self):
        shelf = self.last_shelf()

        if shelf is None:
            raise Exception("No shelf found in '%s'" % self.bzr_root)

        patch = open(shelf, 'r').read()

        print >>sys.stderr, "Reapplying shelved patches",
        message = self.get_shelf_message(patch)
        if message is not None:
            print >>sys.stderr, ' "%s"' % message
        else:
            print >>sys.stderr, ""
        pipe = os.popen('patch -d %s -s -p0' % self.bzr_root, 'w')
        pipe.write(patch)
        pipe.flush()

        if pipe.close() is not None:
            raise Exception("Failed running patch!")

        os.remove(shelf)
        print 'Diff status is now:'
        os.system('bzr diff | diffstat')

        return True

    def shelve(self, all_hunks=False, message=None, revision=None,
             file_list=None):
        cmd = ['diff']
        if revision is not None:
            cmd.extend(['--revision', str(revision[0])])
        if file_list is not None:
            cmd.extend(file_list)
        patches = parse_patches(run_bzr(cmd))
        if not all_hunks:
            try:
                patches = HunkSelector(patches).select()
            except QuitException:
                return False

        if len(patches) == 0:
            print >>sys.stderr, 'Nothing to shelve'
            return True

        shelf = self.next_shelf()
        print >>sys.stderr, "Saving shelved patches to", shelf
        shelf = open(shelf, 'a')
        if message is not None:
            assert '\n' not in message
            shelf.write("# shelf: %s\n" % message)
        for patch in patches:
            shelf.write(str(patch))

        shelf.flush()
        os.fsync(shelf.fileno())
        shelf.close()

        print >>sys.stderr, "Reverting shelved patches"
        pipe = os.popen('patch -d %s -sR -p0' % self.bzr_root, 'w')
        for patch in patches:
            pipe.write(str(patch))
        pipe.flush()

        if pipe.close() is not None:
            raise Exception("Failed running patch!")

        print 'Diff status is now:'
        os.system('bzr diff | diffstat')

        return True

